
require "socket"
require "inifile"

-- configuration
operations = {}
operations.alarm_clear = c8y.config:value('templates/alarm_clear_id', 4000) 
operations.from = c8y.config:value('templates/operation_id', 4001) 
operations.to = operations.from + c8y.config:value('templates/operation_count', 999) - 1
operations.status = c8y.config:value('templates/status', 999)

PORT = c8y.config:value('port')
DEMO = c8y.config:value('demo')

-- not needed now
-- HACK_NAME = 'agent-dac-demo'

varfname = c8y.config:value('dac_volatile')

-----------------------------------------------------------------------------
-- sragent.conf, "plugins" section:
--	lua__dac_machine:
--	{
--		// DAC UDP port.
--		port = 9999
--
--		// File with DAC devices definition.
--		dac_machines = "../conf/dac_machine.ini"
--
--		// File with DAC devices valatile data as IP/port, will be created.
--		dac_volatile = "../conf/dac_volatile.ini"
--
--		// DAC msg/templates IDs
--		templates:
--		{
--			// operations for alarm clear
--			alarm_clear_id = 4000
--			// DAC operations from (range 4001..4999)
--			operation_id = 4001
--			// DAC operations count (range 4001..4999)
--			operation_count = 999
--			// DAC message for set operation status
--			operation_status_id = 999
--		}
--
--		// Demo HACK (define templates in the script)
--		demo = true
--	}
-----------------------------------------------------------------------------
-- /etc/dac_machines.ini
--
-- [soft:sw ver 1.01]
-- xid = 12345
--
-- [sn:10203076]
-- xid = 12346
-- software = sw ver 1.01
-- model = abc
-- class = xyz
--
-----------------------------------------------------------------------------
-- All templates can be defined in function register_demo() if demo = true.
-- And all operations in function add_operations() in DEMO section.
--
-- In this case you can use "@agent-dac-demo" in dac_machine.ini templates 
-- (in software section or device section), but is not needed.
--
-- If you want define templates outside this script just comment HACK_NAME
-- at top of this script and remember that you have to declare operations
-- in add_operations() function.
--
-- Devices are defined in /etc/dac_machines.ini. All defined devices don't
-- need DAC "start" message. All unknown devices can be used just after DAC
-- "start" message.
--
-- The script store last IP and port of DAC devices in "dac_volatile" file
-- (default: /var/lib/sragent/dac_volatile.ini). If IP,port is known and
-- a device is defined in /etc/dac_machines.ini you can send operations to
-- the device without waiting for eny massage from this device.
-----------------------------------------------------------------------------

-- DAC machine object
class "DAC"

-- DAC constructor, just define members
function DAC:DAC()
	self.templates = 0
	self.serial_no = ''
	self.ip = nil
	self.port = nil
	self.device = nil
end


-- template X-ID by software version
templates = {}
-- DAC objected by S/N
dacs = {}


-- create template (by X-Id)
function create_template(entry)
	local xid = entry['xid']
	if not xid then
		return templates[entry['software']]
	end
	xid = tostring(xid)
	if xid:startswith('@') then
		-- hack for demo
		print('Demo template', xid:sub(2))
		return c8y.templates:new(xid:sub(2))
	else
		return c8y.templates:create(xid)
	end
end


-- load DACs form ini file
function load_ini()
	local fname = c8y.config:value('dac_machines')
	if not fname then
		error('No "dac_machines" file in configuration')
	end

	local ini = inifile.parse(fname)
	local varini

	if varfname and file_exists(varfname) then
		varini = inifile.parse(varfname)
	end

	-- load templates (by software version)
	for key,entry in pairs(ini) do
		if key:startswith('soft:') then
			local sw = key:sub(6)
			local tmpl = create_template(entry)
			if not tmpl then
				error('No templates for software "' .. key .. '"')
			end
			templates[sw] = tmpl
		end
	end

	-- load DACs
	for key,entry in pairs(ini) do
		print(key)
		if key:startswith('sn:') then
			local sn = key:sub(4)
			print(sn)
			dacs[sn] = DAC()
			local dac = dacs[sn] 
			dac.serial_no = sn
			dac.soft = entry['software']
			dac.model = entry['model']
			dac.class = entry['class']
			dac.templates = create_template(entry)
			if not dac.templates then
				error('No templates for DAC "' .. key .. '"')
			end
			if varini and varini[key] then
				dac.ip = varini[key].ip
				dac.port = varini[key].port or PORT
			end
		end
	end
	DUMP(dacs)
end


-- define operations
function add_operations()
	c8y.operations:addOperation('c8y_ClearedAlarm', '$.templateId', alarmCleared)
	if DEMO then
		c8y.operations:addOperation('c8y_Relay', '$.id, $.relayState', operation, c8y.operations.ResponseString, 0, 4001)
	end
	c8y.operations:addUnknownOperation(operation)
end

-- LuaAPI: initialize plugin
function init()
	print("DAC proxy initialize at port " .. PORT)
	load_ini()
	getDeviceTmpl = c8y.templates:addRequest('GET,/inventory/managedObjects/%%,,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED,')
	childDeviceResp = c8y.templates:addResponse(',,$.id,$.name,$.type,$.c8y_Hardware.serialNumber')
	add_operations()
end


-- LuaAPI: start running (after main device integrtation)
function start()
	udp = socket.udp()
	udp:setsockname('*', PORT)
	udp:settimeout(0.2)

	scan_children()

	for sn,dac in pairs(dacs) do
		dac.device = c8y.device:childDevice('DAC %%', 'DAC', sn)
		dac.device:update{model='model:' .. dac.model .. ' class:' .. dac.class, swVersion=dac.soft}
	end

	register_demo()
end


function scan_children()
	local resp = getDeviceTmpl:send({c8y.device})
	if resp.ok then
		for i = 1, resp.count do
			local rec = resp:get(i - 1)
			if rec[1] == childDeviceResp.id then
				local id   = rec[3]
				local name = rec[4]
				local typ  = rec[5]
				local sn   = rec[6]
				if dacs[sn] ~= nil then
					dacs[sn].device = c8y.device:new(name, typ, sn)
				end
			end
		end
	end
end


-- register new DAC machine by S/N (msg:100)
function register_dac(d, addr, port)
	-- d: 1=100, 2=s/n, 3=soft, 4=model, 5=class, 6=production_time
	local sn = d[2]
	local soft = d[3]
	-- DAC by S/N
	local dac = dacs[sn] 
	-- if DAC doesn't exist, create one
	if dac == nil then
		dacs[sn] = DAC()
		dac = dacs[sn] 
		dac.templates = templates[soft]
	end
	-- fill DAC with new values
	if varfname and (dac.ip ~= addr or dac.port ~= port) then
		if not varini then varini = {} end
		key = 'sn:'..sn
		if not varini[key] then varini[key] = {} end
		varini[key].ip = addr
		varini[key].port = port
		DUMP('DEBUG: Saving volatile ini', varfname)
		inifile.save(varfname, varini)
	end
	dac.serial_no = sn
	dac.soft = soft
	dac.ip = addr
	dac.port = port
	dac.model = d[4]
	dac.class = d[5]
	dac.device = c8y.device:childDevice('DAC %%', 'DAC', sn)
	dac.device:update{model='model:' .. dac.model .. ' class:' .. dac.class, swVersion=dac.soft}
end


-- remove DAC machine (msg:101)
function remove_dac(sn)
	for i,v in pairs(dacs) do
		-- if v.ip == ip and v.port == port then
		if v.sn == sn then
			-- dacs[i] = nil
			print('I should terminate DAC ' .. sn)
			return
		end
	end
end


-- find DAC by cumulo device ID
function get_dac_by_device_id(id)
	for i,v in pairs(dacs) do
		if v.device and v.device.id == id then
			return dacs[i]
		end
	end
	print('WARNING: no DAC, device.id='..id)
end


--  -- find DAC by IP and port
--  function get_dac_by_ip(ip, port)
--  	for i,v in pairs(dacs) do
--  		if v.ip == ip and v.port == port then
--  			return dacs[i]
--  		end
--  	end
--  	print('WARNING: no DAC, ip='..ip..', port='..port)
--  end


-- find DAC by S/N
function get_dac(sn)
	print(type(sn));
	local dac = dacs[sn]
	if not dac then
		print('WARNING: no DAC, s/n='..sn)
	end
	return dac
end


-- DAC machine message
-- templates arguments is known in soft version
function send(dac, msg, ...)
	if dac == nil then return end
	print('Sending DAC(' .. dac.serial_no .. ') message:' .. msg .. ' with X-Id:' .. tostring(dac.templates.xid))
	if not dac.templates:send({msg, dac.device, ...}) then
		print('Send failed')
	end
end


-- set cumulo DAC machine message
-- templates arguments is known in soft version
function set_status(dac, op, status, failureReason)
	if dac == nil then return end
	print('Sending DAC(' .. dac.serial_no .. ') set status:' .. status .. ' for operation ' .. op .. ', reason: ' .. (failureReason or ''))
	if     status == 'FAILED'     or status == '0' then status = c8y.operations.Failed
	elseif status == 'SUCCESSFUL' or status == '1' then status = c8y.operations.Successful
	else   error('Unknow operation ' .. op .. ' status "' .. status .. '" from DAC(' .. dac.serial_no .. ')')
	end
	if not c8y.operations:setOperationStatus(op, status, failureReason) then
		print('WARNING: Set operation ' .. op .. ' status "' .. status .. '" from DAC(' .. dac.serial_no .. ') failed.')
	end
end


-- LuaAPI: run step
function run()
	errors = {}
	while true do
		-- reveive message
		local d, a, p = udp:receivefrom()
		if d == nil then break end
		-- debug
		print('Got:', d:strip(), a, p)
		-- determine message  ID
		d = d:strip():split(',')
		s,e = pcall(function()
			msg = tonumber(d[1])
			sn = d[2]
		end)
		if s then -- execute message
			if msg == operations.status then set_status(get_dac(sn), unpack(d, 3, #d)) end
			if msg == 100 then register_dac(d, a, p) end
			if msg == 101 then remove_dac(a, p) end
			if msg >= 102 then send(get_dac(sn), msg, unpack(d, 3, #d)) end
		else
			print('Error in run:', e)
			errors[#errors+1] = {a, p, d, e}
		end
	end
	if next(errors) ~= nil then
		error(c8y.dump(errors))
	end
end


-- LuaAPI: stop running
function stop()
	udp:close()
end


-- LuaAPI: unload plugin
function exit()
end

-- Cumulo notification, that alarm is cleared
function alarmCleared(op)
	local record = op.responseRecord
	local alarm  = record[3]
	print('Alarm '..alarm..' cleared in operation '.. op.id)
	dac = get_dac_by_device_id(op.deviceId)
	if dac then
		data = '' .. operations.alarm_clear .. ',' .. alarm
		if dac.ip then
			udp:sendto(data, dac.ip, dac.port)
		else
			DUMP('dac:', dac)
			op:setFailed('DAC device ' .. tostring(dac.serial_no) .. ' has unknown IP')
		end
	else
		DUMP('dac:', dac)
		op:setFailed('Unknown device ' .. op.deviceId)
	end
end

-- Cumulo notification, operation for DAC
function operation(op)
	local record = op.responseRecord
	DUMP('DAC operation', op.id, record)
	local msg  = record[1]
	dac = get_dac_by_device_id(op.deviceId)
	if dac and msg and msg >= operations.from and msg <= operations.to then
		data = '' .. msg .. ',' .. op.id
		for i = 3, #record do
			data = data .. ',' .. record[i]
		end
		if dac.ip then
			udp:sendto(data, dac.ip, dac.port)
		else
			DUMP('dac:', dac)
			op:setFailed('DAC device ' .. tostring(dac.serial_no) .. ' has unknown IP')
		end
	else
		DUMP('dac:', dac)
		op:setFailed('Unknown operation')
	end
end




-- demo hack
function register_demo()
	if DEMO and HACK_NAME and #HACK_NAME then
		DUMP('-x-x-x-x-x-x-x--x-x-x-x-x-x-x-x-x-x-x-x-')
		-- declare all templates here
		xid = c8y.templates:new(HACK_NAME)
		xid:addRequest('POST,/alarm/alarms,application/vnd.com.nsn.cumulocity.alarm+json,application/vnd.com.nsn.cumulocity.alarm+json,%%,NOW UNSIGNED,"{""time"":""%%"",""source"":{""id"":""%%""},""type"":""DacAlarm"",""status"":""ACTIVE"",""severity"":""MINOR"",""text"":""Test alarm only"", ""templateId"":300}"', 300)
		xid:sync()
		DUMP('Demo template X-Id:', xid.xid)
		--c8y.templates:addResponse('$.c8y_Relay,,$.relayState', 4001)
	end
end

