require("tools")
local debugEnabled = c8y.config:value("debug", true)
function debug( ... )
  if debugEnabled then
    print( ... )
  end
end

local templates = {}

local successful = "SUCCESSFUL"
local failed = "FAILED"
local executing = "EXECUTING"

local opStorageDir = c8y.config:value("opStorage", "/opt/deviceShell/opStorage/")
local progressStorageDir = c8y.config:value("progressStorage", "/opt/deviceShell/progressStorage/")
local executor

function ensureDirectoryExists(dir)
  local file = io.popen("mkdir -p " .. dir)
  file:close()
end

function findLinkProfile()
  local result
  for index=1, 6 do
    result = rdbGet("link.profile." .. index .. ".defaultroute")
    if result == "1" then
      debug("FIND LINK PROFILE", index)
      return index
    end
  end
  return nil
end

-- TODO: fix waiting
function wait(time)
  local file = io.popen("rdb wait aaa " .. time)
  local result = file:read("*a")
  file:close()
end

function validateText(text)
  local splitText = text:split_simple(";")
  local action
  for _, currentCommand in ipairs(splitText) do
    action = CommandAction(currentCommand)
    if not action:isValid() then
      return false
    end
  end
  return true
end

function getProgress()
  local file = io.open(progressStorageDir .. "progress")
  if file then
    local progress = file:read("*n")
    file:close()
    return progress
  end
end

function setProgress(value)
  local file = io.open(progressStorageDir .. "progress","w+")
  if value then
    file:write(value)
  end
  file:close()
end

function getCurrentId()
  local file = io.open(progressStorageDir .. "currentId")
  if file then
    local id = file:read("*n")
    file:close()
    return id
  end
end

function setCurrentId(value)
  local file = io.open(progressStorageDir .. "currentId","w+")
  if value then
    file:write(value)
  end
  file:close()
end

function getResult()
  local file = io.open(progressStorageDir .. "result")
  if file then
    local result = file:read("*a")
    file:close()
    return result
  end
end

function setResult(value)
  local file = io.open(progressStorageDir .. "result","w+")
  if value then
    file:write(value)
  end
  file:close()
end

function appendResult(text)
  local file = io.open(progressStorageDir .. "result","a+")
  local previous = file:read("*a")
  debug("APPEND RESULT", text, previous)
  if previous and previous ~= "" then
    file:write("; " .. text)
  else
    file:write(text)
  end
  file:close()
end

function loadOperation(id)
  local operation = {}
  operation.id = id
  local file = io.open(opStorageDir .. id)
  local text = file:read("*l")
  file:close()
  operation.text = text
  local commands = text:split_simple(";")
  local parsedCommands = {}
  print(c8y.dump(commands), type(commands))
  for index, command in ipairs(commands) do
    parsedCommands[index] = CommandAction(command)
  end
  operation.commands = parsedCommands
  return operation
end

function escapeCharacters(string)
  debug("ESCAPE CHARACTERS", string)
  return string:gsub("([ ;&])", "\\%1")
end

function rdbGet(parameter)
  debug("RDB GET", parameter)
  local file = io.popen("rdb get " .. parameter)
  local result = file:read("*l")
  file:close()
  return result
end

function rdbSet(parameter, value)
  debug("RDB SET", parameter, value)
  if type(value) == "string" then
    value = escapeCharacters(value)
  end
  debug("RDB SET AFTER ESCAPE", parameter, value)
  local file = io.popen("rdb set " .. parameter .. " " .. (value or '""'))
  file:close()
end

function saveOperation(id, text)
  local file = io.open(opStorageDir .. id,"w+")
  file:write(text)
  file:close()
end

-------------------------------------------------------------------------------
class("CommandAcceptor")
function CommandAcceptor:operationExecuting(op)
  print('RECEIVED', op.id, op.name, c8y.dump(op:getResponseRecords()))
  local rawText = op:getResponseRecord(1)[3]
  debug(rawText)
  if validateText(rawText) then
    saveOperation(op.id, rawText)
  else
    op:setStatus(c8y.operations.Failed, "Invalid command string")
  end

end
-------------------------------------------------------------------------------
class("CommandAction")
function CommandAction:CommandAction(stringToParse)
  local splitString = stringToParse:split_simple(" ")
  self.type = splitString[1]
  if self.type == "get" then
    self.param = splitString[2]
  elseif self.type == "set" then
    local splitParam = splitString[2]:split_simple("=")
    self.param = splitParam[1]
    self.value = splitParam[2]
    if self.value == '""' then
      self.value = nil
    end
  elseif self.type == "execute" then
    self.command = splitString[2]
  end
  return self
end

function CommandAction:execute()
  self.state = executing
  if self.type == "get" then
    self:_executeGet()
  elseif self.type == "set" then
    self:_executeSet()
  elseif self.type == "execute" then
    self:_executeCommand()
  end
end

function CommandAction:isValid()
  if self.type == "get" then
    if self.param then
      return true
    else
      return false
    end
  elseif self.type == "set" then
    if self.param then
      return true
    else
      return false
    end
  elseif self.type == "execute" then
    if self.command then
      return true
    else
      return false
    end
  end
end

function CommandAction:_executeGet()
  local result = rdbGet(self.param)
  --  maybe some message for not found value on get and even fail the action?
  self.result = result or "No parameter"
  if self.result == "" then
  	self.result = "No value"
  end
  self.state = successful
end

function CommandAction:_executeSet()
  local result = rdbSet(self.param, self.value)
  --  will never fail
  if self.value then
    self.result = "set " .. self.param .. "=" .. self.value
  else
    self.result = "cleared " .. self.param
  end
  self.state = successful
end

function CommandAction:_setPdp(onOff)
  local linkProfile = findLinkProfile()
  if linkProfile then
    rdbSet("link.profile." .. linkProfile .. ".enable", onOff)
    self.state = successful
  end
end

function CommandAction:_executePdpCycle()
  local linkProfile = findLinkProfile()
  if linkProfile then
    rdbSet("link.profile." .. linkProfile .. ".enable", 0)
    wait(5)
    rdbSet("link.profile." .. linkProfile .. ".enable", 1)
    self.state = successful
  end
end

function CommandAction:_executeCommand()
  if self.command == "reboot" then
    io.popen("reboot")
  elseif self.command == "pdpcycle" then
    self:_executePdpCycle()
  elseif self.command == "pdpdown" then
    self:_setPdp(0)
  elseif self.command == "pdpup" then
    self:_setPdp(1)
  else
    self.state = failed
    self.failureReason = ' command "' .. self.command .. '" not supported'
  end
  if self.state == successful then
    self.result = "executed " .. self.command
  end
end

function CommandAction:toString()
  if self.type == "get" then
    return "get " .. self.param
  elseif self.type == "set" then
    return "set " .. self.param .. "=" .. (self.value or "")
  elseif self.type == "execute" then
    return "execute " .. self.command
  end
end

-------------------------------------------------------------------------------
class("OperationExecutor")
function OperationExecutor:run()
  self.running = true
  local currentId = getCurrentId()
  if currentId then
    self:resumeOperation(currentId)
  else
    self:startNextOperation()
  end

  if self.op then
    os.remove(opStorageDir .. self.op.id)
    setResult()
    setCurrentId()
    self.op = nil
  end
  self.running = false
end

function OperationExecutor:generateResponse()
  debug("GENERATE RESPONSE", getResult(), self.op.text)
  return '{"result":"' .. getResult() .. '","text":"' .. self.op.text .. '"}'
end

function OperationExecutor:resumeOperation(id)
  local progress = getProgress()
  local operation = loadOperation(id)
  self.op = operation
  local commands = operation.commands

  local currentCommand = commands[progress]
  appendResult("executed " .. currentCommand.command)
  progress = progress + 1

  self:executeCommands(progress, commands)
end

function OperationExecutor:startNextOperation()
  local ls = io.popen("ls " .. opStorageDir)
  local id = ls:read("*n")
  ls:close()
  if id then
    debug("NEXT", id)
    setCurrentId(id)
    local operation = loadOperation(id)
    self.op = operation
    self:executeCommands(1, operation.commands)
  end
end

function OperationExecutor:executeCommands(progress, commands)
  debug("EXECUTE COMMANDS", progress, c8y.dump(commands), #commands)
  local currentCommand
  for i=progress, #commands do
    setProgress(i)
    currentCommand = commands[i]
    debug("CURRENT COMMAND", c8y.dump(currentCommand))
    currentCommand:execute()
    while (not currentCommand.state) or currentCommand.state == executing do
      wait(1)
    end
    if currentCommand.state == failed then
      c8y.operations:setOperationStatus(self.op.id, c8y.operations.Failed, "Executing command " .. currentCommand:toString() .. " failed" .. (currentCommand.failureReason or ""))
      return
    end
    appendResult(currentCommand.result)
  end
  local id = self.op.id
  local generatedResponse = self:generateResponse()
  local response = templates.operationResponse:send({id, "SUCCESSFUL", generatedResponse, JsonEscape=false})

  c8y.timer.start( 5, function(timer)
    if response.httpCode < 0 then
      debug("RESPONSE AFTER SEND", response.ok, response.httpCode)
      response = templates.operationResponse:send({id, "SUCCESSFUL", generatedResponse, JsonEscape=false})
    else
      timer:stop()
    end
  end)
  c8y.operations:opeartionHandled(self.op.id)
end
-------------------------------------------------------------------------------

function init()
  ensureDirectoryExists(opStorageDir)
  ensureDirectoryExists(progressStorageDir)
  executor = OperationExecutor()
  local commandHandler = CommandAcceptor()
  c8y.operations:addOperation("c8y_Command", "$.text", commandHandler)
  templates.operationResponse = c8y.templates:addRequest('PUT,/devicecontrol/operations/%%,application/vnd.com.nsn.cumulocity.operation+json,,%%,UNSIGNED STRING STRING,"{""status"":""%%"",""c8y_Command"":%%}"')
end

function run()
  if not executor.running then
    executor:run()
  end
end
