
-- http://www.modbus.org/docs/Modbus_Messaging_Implementation_Guide_V1_0b.pdf
-- http://www.modbus.org/docs/Modbus_Application_Protocol_V1_1b.pdf

--require "socket"
require "struct"
require "modbus"
ModBus = modbus.ModBus
ModBusMessage = modbus.ModBusMessage

function init()
	test()
	--modbus = ModBus{}
	modbus = {}
	c8y.operations:addOperation('c8y_ModbusDevice', '$.ipAddress,$.address,$.name,$.type', modbusDeviceOperation)
	-- c8y.operations:addOperation('c8y_SetRegister',
	-- 	'$c8y_SetRegister.register,$c8y_SetRegister.address,$c8y_SetRegister.value',
	-- 	setRegisterOperation)
	c8y.operations:addOperation('c8y_SetRegister', '$.address,$.register,$.value', setRegisterOperation)
	c8y.operations:addOperation('c8y_SetCoil', '$.address,$.coil,$.value', setCoilOperation)
	c8y.operations:addOperation('c8y_ModbusConfiguration', '$.pollingRate,$.transmitRate', setModbusConfiguration)
	--devices = {}   - now in modbus table
	deviceTypes = {}
	templates={}
	templates.updateParentModbusConfiguration = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED UNSIGNED UNSIGNED,"{""c8y_ModbusConfiguration"":{""protocol"": ""TCP"", ""pollingRate"": %%, ""transmitRate"": %%}}"')
	templates.update = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED UNSIGNED STRING STRING,"{""c8y_ModbusDevice"":{""address"":%%,""type"":""%%"",""ipAddress"":""%%""}}"')
	templates.getDevice = c8y.templates:addRequest('GET,/inventory/managedObjects/%%,,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED,')
	templates.getOperations = c8y.templates:addRequest('GET,/devicecontrol/operations?deviceId=%%&status=%%&pageSize=100,,,%%,UNSIGNED STRING,')
	templates.parentModbusResponse = c8y.templates:addResponse('$.c8y_ModbusConfiguration,,$.pollingRate,$.transmitRate')
	--templates.updateOneCoil = c8y.templates:addRequest('GET,/inventory/managedObjects/%%,,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED,"{""c8y_SetCoil"":{""coil"":%%,""value"":%%,""address"":%%}"')
	templates.opResponse = c8y.templates:addResponse('$.operations,,$.id,$.status,$.deviceId')

	DeviceType.addTemplates(templates)

	interval = {}
	interval.pollingRate = c8y.config:value('read_interval', 15)
	interval.transmitRate = c8y.config:value('measurement_interval', 30)
	interval.readTimeout = c8y.config:value('read_timeout', 30)
	offset = c8y.config:value('address_offset', 0)
end

function start()
	scanModBusDevices()
	readRegisters()
end

function run()
	-- for _, m in pairs(modbus) do
	-- 	while m:receive() ~= nil do
	-- 	end
	-- end
end

function createModBus(args)
	local m = ModBus(args)
	m.alarms = {}
	m:setConnectionTimeout(15)
	modbus[args.deviceId] = m
	return m
end

function scanModBusDevices()
	print('Scanning ModBus devices')

	-- scan modbus intervals and children devices
	local hasModbusFragment = false
	local resp = templates.getDevice:send({c8y.device})
	if resp.ok then
		for i = 1, resp.count do
			local rec = resp:get(i - 1)
			if rec[1] == templates.parentModbusResponse.id then
				local pollingRate  = rec[3]
				local transmitRate = rec[4]
				hasModbusFragment = true
				interval.pollingRate = math.max(pollingRate, 1)
				interval.transmitRate = math.max(transmitRate, 1)
				if readRegistersTimer then readRegistersTimer.interval = interval.pollingRate end
				if sendMeasurementTimer then sendMeasurementTimer.interval = interval.transmitRate end
			elseif rec[1] == templates.childDevice.id then
				local devId   = rec[3]
				local devName = rec[4]
				local devresp = templates.getDevice:send({devId})
				if devresp.ok then
					for j = 1, devresp.count do
						local devrec = devresp:get(j - 1)
						if devrec[1] == templates.modbusDevice.id then
							local ip   = devrec[3]
							local unit = devrec[4]
							local typ  = devrec[5]
							print('Found ModBus device '..ip..':'..unit..' named '..devName)
							if deviceTypes[typ] == nil then deviceTypes[typ] = DeviceType(typ) end
							createModBus{server=ip, unit=unit, deviceId=devId, name=devName, deviceType=deviceTypes[typ]}
						end
					end
				end
				local opresp = templates.getOperations:send({devId,'PENDING'})
				if opresp.ok then
					for _,rec in pairs(opresp:getRecords(templates.opResponse.id)) do
						local id = rec[3]
						c8y.operations:setOperationStatus(id, c8y.operations.Executing)
					end
				end
				local opresp = templates.getOperations:send({devId,'EXECUTING'})
				if opresp.ok then
					for _,rec in pairs(opresp:getRecords(templates.opResponse.id)) do
						local id = rec[3]
						c8y.operations:setOperationStatus(id, c8y.operations.Failed)
					end
				end
			end
		end
	end
	--DUMP('-x-x-x-x-x-x-x-x-x-x-x-x-   hasModbusFragment', hasModbusFragment, interval, 'tmplID', templates.parentModbusResponse.id)
	scanDeviceTypes()
	if not hasModbusFragment then
		templates.updateParentModbusConfiguration:send({c8y.device, interval.pollingRate, interval.transmitRate})
	end
end

-- scan types
function scanDeviceTypes()
	for _, t in pairs(deviceTypes) do
		t:update(templates)
	end
end



-----------------------------------------------------------------------------
-- ModBus Device Type (device description)
class "DeviceType"

-- New device type
function DeviceType:DeviceType(typ)
	self.typ = typ
	self.id = typ:gsub('.*/','')
	self.id = tonumber(self.id)
	self.xid = self.id
	self.registers = {}
	self.coils = {}
	self.measurements = {}
	self.alarmsByTmpl = {}
	self.alarmsByAddr = {}
	self.template = c8y.templates:create(self.xid)
	self.useServerTime = false
	self.updated = false
end

-- Send template with X-Id from this device type
function DeviceType:send(args)
	return self.template:send(args)
end

-- Add templates for discover device type
-- static function
function DeviceType.addTemplates(templates)
	templates.getTypeDevice = c8y.templates:addRequest('GET,/inventory/managedObjects/%%,,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED,')
	templates.childDevice = c8y.templates:addResponse('$.childDevices.references,,$.managedObject.id,$.managedObject.name')
	templates.modbusDevice = c8y.templates:addResponse('$.c8y_ModbusDevice,,$.ipAddress,$.address,$.type')
	templates.modbusDeviceType = c8y.templates:addResponse(',$.c8y_useServerTime,$.c8y_useServerTime')
	templates.registersDef = c8y.templates:addResponse('$.c8y_Registers,$.statusMapping,$.number,$.multiplier,$.divisor,$.decimalPlaces')
	templates.coilsDef = c8y.templates:addResponse('$.c8y_Coils,$.statusMapping,$.number')
	templates.regStatus = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED INTEGER INTEGER,"{""c8y_RegisterStatus"":{""register2"":%%,""register1"":""%%""}}"')
	templates.measurementDef = c8y.templates:addResponse('$.c8y_Registers,$.measurementMapping,$.measurementMapping.sendMeasurementTemplate,$.number')
	templates.alaramsDef = c8y.templates:addResponse('$.c8y_Coils,$.alarmMapping,$.alarmMapping.raiseAlarmTemplate,$.number,$alarmMapping.type,$.alarmMapping.severity,$.alarmMapping.text')
end

-- Update device type description
function DeviceType:update(templates)
	if not self.updated then
		local resp = templates.getTypeDevice:send({self.id})
		if resp.ok then
			for i = 1, resp.count do
				local rec = resp:get(i - 1)
				-- global settings
				if rec[1] == templates.modbusDeviceType.id then
					self.useServerTime = rec[3]
				end
				-- registers
				if rec[1] == templates.registersDef.id then
					local num  = rec[3] - offset
					local mul  = rec[4]
					local div  = rec[5]
					local prec = rec[6]
					if not mul or mul == 0 then mul = 1; end
					if not div or div == 0 then div = 1; end
					self.registers[num] = DeviceRegister{deviceType=self, addr=num, mul=mul, div=div, prec=prec}
				end
				-- measurements (registers)
				if rec[1] == templates.measurementDef.id then
					local tmpl = rec[3]
					local num  = rec[4] - offset
					if self.measurements[tmpl] == nil then self.measurements[tmpl] = {} end
					local measurements = self.measurements[tmpl]
					measurements[#measurements+1] = DeviceMeasurement{deviceType=self, tmpl=tmpl, addr=num}
				end
				-- coils
				if rec[1] == templates.coilsDef.id then
					local num = rec[3] - offset
					--DUMP('XXX', 'new coil', num)
					self.coils[#self.coils+1] = num
				end
				-- alarms (colis)
				--DUMP('XXX', 'alaramsDef', templates.alaramsDef.id)
				if rec[1] == templates.alaramsDef.id then
					local tmpl = rec[3]
					local num  = rec[4] - offset
					local typ  = rec[5]
					local svr  = rec[6]
					local text = rec[7]
					--DUMP('XXX', {deviceType=self, tmpl=tmpl, addr=num, alarmType=typ, severity=srv, text=text})
					local alarm = DeviceAlarm{deviceType=self, tmpl=tmpl, addr=num, alarmType=typ, severity=srv, text=text}
					if self.alarmsByTmpl[tmpl] == nil then self.alarmsByTmpl[tmpl] = {} end
					local alarms = self.alarmsByTmpl[tmpl]
					alarms[#alarms+1] = alarm
					self.alarmsByAddr[num] = alarm
					
					local exists
          for _, coil in ipairs(self.coils) do
            if coil == num then
              exists = true
              break
            end
          end
          if not exists then
            self.coils[#self.coils+1] = num
            alarm.alarmOnly = true
          end
				end
			end
			-- in coil/register number order  (templates: 100, 101)
			table.sort(self.coils)
			for _,m in pairs(self.measurements) do table.sort(m, function(a,b) return a.addr < b.addr end) end
		end
	end
end

-- Get sorted list of coils adresses
function DeviceType:getCoilAddrs()
	-- local out = {}
	-- for k,_ in pairs(self.coils) do out[#out+1] = k end
	-- table.sort(out)
	-- return out
	return self.coils
end

-- Get sorted list of register adresses
function DeviceType:getRegisterAddrs()
	local regs = {}
	for k,_ in pairs(self.registers) do regs[#regs+1] = k end
	table.sort(regs)
	return regs
end


-----------------------------------------------------------------------------
-- ModBus Device Register
class "DeviceRegister"

function DeviceRegister:DeviceRegister(args)
	self.addr = args.addr
	self.mul = args.mul or 1
	self.div = args.div or 1
	self.prec = args.prec or 0
	self.deviceType = args.deviceType
end

-- -- send measurement
-- function DeviceRegister:send(deviceId, val)
-- 	args = {self.tmpl, deviceId}
-- 	if not self.useServerTime then args[#args+1] = os.date() end
-- 	args[#args+1] = val
-- 	return self.deviceType:send(args)
-- end


-----------------------------------------------------------------------------
-- ModBus Device Measurements
class "DeviceMeasurement"

function DeviceMeasurement:DeviceMeasurement(args)
	self.tmpl = args.tmpl
	self.addr = args.addr
	self.mul = args.mul or 1
	self.div = args.div or 1
	self.prec = args.prec or 0
	self.deviceType = args.deviceType
end


-----------------------------------------------------------------------------
-- ModBus Device Alarm
class "DeviceAlarm"

function DeviceAlarm:DeviceAlarm(args)
	self.tmpl = args.tmpl
	self.addr = args.addr
	self.alarmType = args.alarmType
	self.severity = args.severity
	self.text = args.text
	self.deviceType = args.deviceType
end


-----------------------------------------------------------------------------
-- The c8y_SetRegister operation handler
function setModbusConfiguration(op)
	local rec = op.responseRecord
	local pollingRate  = rec[3]
	local transmitRate = rec[4]
	interval.pollingRate = math.max(pollingRate, 1)
	interval.transmitRate = math.max(transmitRate, 1)
	if readRegistersTimer then readRegistersTimer.interval = interval.pollingRate end
	if sendMeasurementTimer then sendMeasurementTimer.interval = interval.transmitRate end
	templates.updateParentModbusConfiguration:send({c8y.device, interval.pollingRate, interval.transmitRate})
	op.successful = true
	return true
end


-----------------------------------------------------------------------------
-- The c8y_ModbusDevice operation handler
function modbusDeviceOperation(op)
	local record = op.responseRecord
	local ip   = record[3]
	local unit = record[4]
	local name = record[5]
	local typ  = record[6]
	local device = c8y.device:childDevice(name, 'c8y_ModbusDevice', 'modbus:'..ip..':'..unit)
	print('Adding child device', device.id)
	templates.update:send({device, unit, typ, ip})
	if deviceTypes[typ] == nil then deviceTypes[typ] = DeviceType(typ) end
	createModBus{server=ip, unit=unit, deviceId=device.id, name=name, deviceType=deviceTypes[typ]}
	op.successful = true
	scanDeviceTypes()
	return true
end


-----------------------------------------------------------------------------
-- The c8y_SetCoil operation handler
function setCoilOperation(op)
	--DUMP('coilStatE:', op.id, op.name, op.records, op.deviceId)
	local record = op.responseRecord
	local addr = record[3]
	local coil = record[4] - offset
	local val  = record[5]
	local msg = modbus[op.deviceId]:setCoil(coil, val, function(msg)
		print('finished: op.id='..op.id..', modbus.id='..msg.id..', coil='..coil..', addr='..addr..', val='..val)
		if msg.ok then
			op.successful = true
		else
			op:setFailed('ModBus error: ' .. msg:errorString())
		end
	end)
	return true
end


-----------------------------------------------------------------------------
-- The c8y_SetRegister operation handler
function setRegisterOperation(op)
	--DUMP('regE:', op.id, op.name, op.records, op.deviceId)
	local m = modbus[op.deviceId]
	local record = op.responseRecord
	local addr = record[3]
	local reg  = record[4] - offset
	local val  = record[5]
	local r = m.deviceType.registers[reg]
	if not r or r.mul == 0 then
		val = 0
	else
		val = val * r.div / r.mul
	end
	local msg = m:setRegister(reg, val, function(msg)
		print('finished: op.id='..op.id..', modbus.id='..msg.id)
		if msg.ok then
			op.successful = true
		else
			op:setFailed('ModBus error: ' .. msg:errorString())
		end
	end)
	return true
end


-----------------------------------------------------------------------------
-- Reading registers
function readRegisters()
	if readRegistersTimer == nil then
		readRegistersTimer = c8y.timer.start(interval.pollingRate, readRegisters)
	end
	if next(modbus) == nil then return end
	print('Updating coils and registers status. Interval '..readRegistersTimer.interval)
	for _,m in pairs(modbus) do
		if m.modbus_reading then
			print('ModBus device' .. m.name .. ' updating coils and registers status in progress - ignoring')
		else
			local function failed()
				print('ModBus device' .. m.name .. ' register and coils reading FAILED')
				m.modbus_reading = false
			end
			m.modbus_reading = c8y.timer.singleShot(interval.readTimeout, failed)
			print('Reading data from ' .. m.name)
			-- on registers read finished
			local function registersCallback(values)
				DUMP(m.name, 'REGS:', values)
				if values == nil then return failed() end
				local regs = m.deviceType:getRegisterAddrs()
				for i, addr in ipairs(regs) do
					local reg = m.deviceType.registers[addr]
					if reg.div == 0 then
						values[i+2] = 0
					else
						values[i+2] = values[i+2] * reg.mul / reg.div
					end
				end
				DUMP(m.name, 'Regs:', values)
				m:sendTemplate(values)
				m.lastRegisters = values
				if m.modbus_reading then
					m.modbus_reading:stop()
				end
				m.modbus_reading = false
				if sendMeasurementTimer == nil then sendMeasurement() end
				collectgarbage()  -- remove timers
				return true
			end
			-- on coils read finished
			local function coilsCallback(values)
				DUMP(m.name, 'COILS:', values)
				if values == nil then return failed() end
				local addrs = m.deviceType:getCoilAddrs()
				--DUMP('XXX','coils:',addrs)
				local valuesWithoutAlarmOnly = {values[1], values[2]}
				local skipcount = 0
				for i,a in ipairs(addrs) do
					local alarm = m.deviceType.alarmsByAddr[a]   -- alarms for this coil?
					--DUMP('XXX:alarm:', i, a, values, alarm, values[i+2], m.alarms[a])
					if alarm ~= nil then   -- yes, we have alarm definition for this coil
						if values[i+2] and m.alarms[a] == nil then  --  0 -> 1, new alaram
							local t = c8y.strftime()
							print(m.name, 'Coil '..a..' alarm at '..t)
							local resp = m:sendTemplate{alarm.tmpl, m.deviceId, t}
							if resp.ok then
								m.alarms[a] = true
							else
								print(m.name, 'WARINING: Can not send alarm '..alarm.tmpl)
							end
						elseif not values[i+2] and m.alarms[a] ~= nil then  --  1 -> 0
							m.alarms[a] = nil
						end
						if not alarm.alarmOnly then
              valuesWithoutAlarmOnly[#valuesWithoutAlarmOnly + 1] = values[i+2]
            end
          else
            valuesWithoutAlarmOnly[#valuesWithoutAlarmOnly + 1] = values[i+2]
					end
				end
				m:sendTemplate(valuesWithoutAlarmOnly)
			end
			m:getCoil{addr=m.deviceType.coils, callback=coilsCallback, values={100, m.deviceId}}
			m:getHoldingRegister{addr=m.deviceType:getRegisterAddrs(), callback=registersCallback, values={101, m.deviceId}}
		end
	end
end

-- make table indexed by addr
function makeAddrIndexed(addrs, values, skip)
	skip = skip or 0
	assert(#addrs == #values - skip)
	local out = {}
	for i = 1, #addrs do
		out[addrs[i]] = values[i + skip]
	end
	DUMP('m', addrs, values, skip, out)
	return out
end

-- Send measurement
function sendMeasurement()
	if sendMeasurementTimer == nil then
		sendMeasurementTimer = c8y.timer.start(interval.transmitRate, sendMeasurement)
	end
	if next(modbus) == nil then return end
	print('Send measurement. Interval '..sendMeasurementTimer.interval)
	for _,m in pairs(modbus) do
		if m.lastRegisters then
			-- get last read values
			local last = makeAddrIndexed(m.deviceType:getRegisterAddrs(), m.lastRegisters, 2)  -- skip tmplID, devID
			m.lastRegisters = nil
			-- iterate by all templates (measurements types)
			for tmpl,list in pairs(m.deviceType.measurements) do
				local measurements = { tmpl, m.deviceId }
				if not m.deviceType.useServerTime then measurements[#measurements+1] = c8y.strftime() end
				-- convert 
				for _,s in ipairs(list) do
					local v = last[s.addr]
					if v then
						measurements[#measurements + 1] = v
					else
						print('WANRING: no measurements for addr '..s.addr)
					end
				end
				m:sendTemplate(measurements)
			end
		end
	end
end

-- Send alarams
function sendAlarams()
end


-----------------------------------------------------------------------------
--
function usleep(us)
	DUMP('usleep', us)
	os.execute('usleep '..us)
end
--


function test()
end

