
--
-- SAVE:
--   dbcfg_export -o /tmp/$filename -p "$pw"
--   tar -C /usr/local/cdcs -zcvf - ipsec.d openvpn-keys ssh-hostkeys | openssl des3 -salt -k "$pw" | dd of=/tmp/vpn.des3
--   tar -zcvf $filename.tar.gz $filename vpn.des3
--
--
-- RESTORE:
--   install_file "CONFIG_FILE"
--

local templates = {}
local opfile = '/var/lib/sragent/ntc_config.op'

local debug = c8y.config:boolValue("debug") and print or function() end


function init()
	c8y.operations:addOperation('c8y_Configuration', '$.config', write_config)
	c8y.operations:addOperation('c8y_SendConfiguration', '', read_config)

	templates.update = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED STRING,"{""c8y_Configuration"":{""config"":""%%""}}"')

end

-- Write, tar config and install
function write_config(op)
	local fname = string.format('NTC-6000_Settings_%06d.cfg', math.random(999999))
	local f = io.open('/tmp/'..fname, 'w+')
	f:write(op.record[3])
	f:close()
	local openvpnkeys = file_entry_exists('/usr/local/cdcs/openvpn-keys') and 'openvpn-keys ' or ''
	local sshhostkeys = file_entry_exists('/usr/local/cdcs/ssh-hostkeys') and 'ssh-hostkeys ' or ''
	local c = string.format('cd /tmp && tar -C /usr/local/cdcs -zcvf - ipsec.d ' .. openvpnkeys .. sshhostkeys .. '| openssl des3 -salt -k "" | dd of=/tmp/vpn.des3 && tar czf /opt/cdcs/upload/%s.tar.gz %s vpn.des3 && install_file /opt/cdcs/upload/%s.tar.gz', fname, fname, fname)
	debug( 'conf cmd:', c )
	local f = io.popen(c, 'r')
	local c = f:read('*all')
	debug( 'install_file:', c )
	os.remove('/tmp/'..fname)
	os.remove('/tmp/vpn.des3')
	if c:find('is successful') then
		if c:find('reboot') then
			local f = io.open(opfile, 'w')
			if f then
				-- postpone setting successful after reboot
				f:write(tostring(op.id))
				f:close();
				-- reboot
				os.execute('rdb set service.system.reset 1')
			else
				op:setFailed('install_file successful (need reboot) but  '..c)
			end
		else
			op.successful = true
		end
	else
		op:setFailed('install_file failed: '..c)
	end
end

function send_config()
	local f = io.popen('dbcfg_export -p ""', 'r')
	local c = f:read('*all')
	--DUMP( c )
	templates.update:send({c8y.device, c})
end

function read_config(op)
	send_config()
	op.successful = true
end

function start()
	-- set successful afrer reboot
	if file_exists(opfile) then
		local f = io.open(opfile)
		local op = f:read('*n')
		f:close()
		c8y.operations:setOperationStatus(op, c8y.operations.Successful)
		os.remove(opfile)
	end
	-- check if just registred
	if c8y.device.registered then
		send_config()
	end
end

