require("tools")

local interval
local templates = {}
local timer
function init()
  templates.sendAll = c8y.templates:addRequest('POST,/measurement/measurements,application/vnd.com.nsn.cumulocity.measurement+json,,%%,NUMBER NUMBER NUMBER NOW STRING,"{""c8y_SignalStrength"":{""rssi"":{""value"":%%,""unit"":""dBm""},""ECN0"":{""value"": %%},""RCSP"":{""value"":%%}},""time"":""%%"",""source"":{ ""id"":""%%""},""type"":""c8y_SignalMeasurement""}"')
  templates.sendWithoutEcn0 = c8y.templates:addRequest('POST,/measurement/measurements,application/vnd.com.nsn.cumulocity.measurement+json,,%%,NUMBER NUMBER NOW STRING,"{""c8y_SignalStrength"":{""rssi"":{""value"":%%,""unit"":""dBm""},""RCSP"":{""value"":%%}},""time"":""%%"",""source"":{ ""id"":""%%""},""type"":""c8y_SignalMeasurement""}"')
end

function getRdb(parameter)
  local file = io.popen("rdb get " .. parameter)
  local value = file:read("*n")
  file:close()
  return value
end

function readRssi()
  local rssi = getRdb("wwan.0.radio.information.signal_strength")
  return rssi
end

function readEcn0()
  local valid = getRdb("wwan.0.system_network_status.ECN0_valid")
  if valid == 1 then
    local ecn0 = getRdb("wwan.0.system_network_status.ECN0s0")
    return ecn0
  end
  return nil
end

function readRcsp()
  local rcsp = getRdb("wwan.0.system_network_status.RSCPs0")
  return rcsp
end

function sendMeasurements()
  local rssi = readRssi()
  local ecn0 = readEcn0()
  local rcsp = readRcsp()
  if rssi and rcsp then
    local response
    if ecn0 then
      response = templates.sendAll:send({rssi, ecn0, rcsp, c8y.device})
    else
      response = templates.sendWithoutEcn0:send({rssi, rcsp, c8y.device})
    end
    return response.httpCode == 200
  end
end

function safelySendMeasurements()
  local timer = c8y.timer.start(5, function(timer)
    local result = sendMeasurements()
    if result then
      timer:stop()
    end
  end)
end

function start()
  -- measurement sending timer
  local sendMeasurementsTimer = c8y.timer.new(function()
    sendMeasurements()
  end)
  c8y.timer.start(10, function()
    local n = tonumber(c8y.config:value("interval", 0))
    if n ~= interval then
      print('interval changed '..(interval or "nil") ..' -> '..(n or "nil"))
      if not n or n == 0 then
        print("new interval is " .. (n or "nil") .. ", stopping measurements timer")
        sendMeasurementsTimer:stop()
      else
        if sendMeasurementsTimer.active then
          print("just changing interval")
          sendMeasurementsTimer.interval = n
        else
          print("timer not active, starting measuremnts timer")
          sendMeasurementsTimer:start(n)
        end
      end
      interval = n
    end
  end)
  safelySendMeasurements()
end
