local templates = {}
local c8y_prefix = "service.cumulocity."

class "ReportHandler"
function ReportHandler:operationExecuting(op)
   local sig = readSignal()
   local v1 = readGPIO("xaux1")
   local v2 = readGPIO("xaux2")
   templates.report:send{sig, c8y.device}
   if v1 == 0 or v2 == 0 then
      result = templates.query:send{c8y.device}
   end
   if v1 == 0 then
      for i=result.count,1,-1 do
         local id = result:getRecord(i)[3]
         local type = result:getRecord(i)[4]
         if type == "gpio1" then
            templates.clear:send{id}
            break
         end
      end
   else
      local text = readAlarmText("gpio.1", v1)
      local severity = readAlarmSeverity("gpio.1")
      templates.alarm:send{"gpio1", text, c8y.device, severity}
   end
   if v2 == 0 then
      for i=result.count,1,-1 do
         local id = result:getRecord(i)[3]
         local type = result:getRecord(i)[4]
         if type == "gpio2" then
            templates.clear:send{id}
            break
         end
      end
   else
      local text = readAlarmText("gpio.2", v2)
      local severity = readAlarmSeverity("gpio.2")
      templates.alarm:send{"gpio2", text, c8y.device, severity}
   end
   return op:setStatus(op.Successful)
end


handler = ReportHandler()

function init()
   tc = c8y.templates:new()
   templates.report = tc:addRequest('POST,/measurement/measurements,application/vnd.com.nsn.cumulocity.measurement+json,,%%,NUMBER NOW UNSIGNED,"{""c8y_SignalStrength"":{""rssi"":{""value"":%%,""unit"":""dBm""}},""time"":""%%"",""type"":""c8y_SignalMeasurement"",""source"":{""id"":""%%""}}"')
   templates.alarm = tc:addRequest('POST,/alarm/alarms,application/json,,%%,STRING STRING NOW UNSIGNED STRING,"{""type"":""%%"", ""text"":""%%"", ""time"":""%%"",""source"":{""id"":""%%""},""severity"":""%%"",""status"":""ACTIVE""}"')
   templates.query = tc:addRequest('GET,/alarm/alarms?source=%%&status=ACTIVE,,,%%,UNSIGNED,')
   templates.clear = tc:addRequest('PUT,/alarm/alarms/%%,application/json,,%%,UNSIGNED,"{""status"":""CLEARED""}"')
   templates.response = tc:addResponse('$.alarms,,$.id,$.type')
   c8y.operations:addOperation('c8y_MeasurementRequestOperation', '', handler)
end


function getRdb(parameter, mode)
   local file = io.popen("rdb get " .. parameter)
   local value = file:read(mode)
   file:close()
   return value
end


function readAlarmText(pin, value)
   local text = getRdb(c8y_prefix .. pin .. '.alarm.text', "*l")
   if text and #text > 0 then
      return text
   elseif value == 1 then
      return pin .. ' is active'
   else
      return pin .. ' is inactive'
   end
end


function readAlarmSeverity(pin)
   local text = getRdb(c8y_prefix .. pin .. '.alarm.severity', '*l')
   if text then
      return text
   else
      return 'MAJOR'
   end
end


function readSignal()
   local signal = getRdb("wwan.0.radio.information.signal_strength", "*n")
   if signal then return signal end

   signal = getRdb("wwan.0.system_network_status.RSCPs0", "*n")
   if signal then return -signal end

   local valid = getRdb("wwan.0.system_network_status.ECN0_valid", "*n")
   if valid == 1 then
      signal = getRdb("wwan.0.system_network_status.ECN0s0", "*n")
      return signal
   end
   return 0
end


function readGPIO(pin)
   local prefix = "sys.sensors.io."
   local mode = getRdb(prefix .. pin .. ".mode", "*l")
   local value = 0
   if mode == "analogue_input" then
      local pullup = getRdb("sys.sensors.iocfg.pull_up_voltage", "*n")
      if pullup then
         local a_in = getRdb(prefix .. pin .. ".adc", "*n")
         if a_in+1 > pullup then value = 1 end
      end
   elseif mode == "digital_input" or mode == "virtual_digital_input" then
         value = getRdb(prefix .. pin .. ".d_in", "*n")
   elseif mode == "digital_output" then
      value = getRdb(prefix .. pin .. ".d_out", "*n")
   end
   return value
end
