
require "inifile"

local host = c8y.config:stringValue('bind', 'localhost')
local port = c8y.config:integerValue('port', 1331)
local inifname = c8y.config:value('alarms')

local alarms
local templates = {}

-- initialize module
function init()
	print('host:'..(host or '') .. ', port:'..(port or '') .. ', alarms:'..(inifname or ''))
	if not inifname then
		error('No "alarms" file in configuration')
	end
	alarms = inifile.parse(inifname)
	udp = c8y.net:udp()
	udp:bind(host, port)
	udp.onReceive = receive_alarm
	templates.id = c8y.templates:new()
	templates.alarm = templates.id:addRequest('POST,/alarm/alarms,application/json,,%%,STRING STRING DATE UNSIGNED STRING,"{""type"":""%%"", ""text"":""%%"", ""time"":""%%"",""source"":{""id"":""%%""},""severity"":""%%"",""status"":""ACTIVE""}"')
end



function receive_alarm(s)
	local d = s:recv()
	local id, date, time, dev, msg = d:match('%[EVENT#([^]]+)%] *([-0-9]+) +([0-9:]+) +([^:]-) *: *(.*) *\r?\n.*')
	print('UDP Alarm #'..id..' '..msg)
	--DUMP('DATA:  ', id, date, time, dev, msg)
	if id then
		local timestamp = date .. 'T' .. time .. c8y.strftime():sub(-5)  -- '%FT%T%z'
		local glob = alarms.alarm or {}
		local sect = alarms['alarm.'..id] or {}
		local aname = sect.name or glob.name or 'Unknown alarm #'..id
		local atype = sect.type or glob.type or aname
		local severity = sect.severity or glob.severity or 'MAJOR'
		--DUMP('ALARM: ', id, timestamp, dev, msg, aname, atype, severity)
		if not atype then atype = 'Unknown NetComm alarm' end
		templates.alarm:send{atype, dev..': '..msg, timestamp, c8y.device.id, severity}
	end
end



-- 1::Unit powered up
-- 2::Unit rebooted
-- 3::Link status change
-- 4::WWAN IP address change
-- 5::WWAN Registration change
-- 6::WWAN Cell ID change
-- 7::WWAN technology change
-- 8::Number of connected Ethernet interfaces change
-- 9::Power source change
-- 10::Web UI login failure
-- 11::SD card status change
-- 12::WAN failover instance occurred
-- 13::Number of connected WiFi clients change
-- 14::Monitored WiFi clients status change
-- 15::Digital input change
-- 16::Analog input threshold
-- 17::Digital output change

